"""
CMake dependency graph.
Requires graphviz installed.

Requires user to first invoke "cmake -B build --graphviz=gfx/block.dot"
"""

import shutil
import subprocess
import argparse
from pathlib import Path


def main():
    p = argparse.ArgumentParser(description="convert .dot graph to SVG or PNG")
    p.add_argument("path", help="Path to CMake gfx/ directory from cmake -B build --graphviz=gfx/block.dot")
    p.add_argument("format", help="output format", choices=["svg", "png"], default="svg", nargs="?")
    P = p.parse_args()

    fmt = P.format

    dot = shutil.which("dot")
    if not dot:
        raise FileNotFoundError("GraphViz Dot program not available.")

    path = Path(P.path).expanduser()
    if not path.is_dir():
        raise NotADirectoryError(path)

    try:
        name_file = next(path.glob("*.dot"))
    except StopIteration:
        raise FileNotFoundError(f"No .dot files in {path}")

    name_pat = name_file.name

    for file in path.glob(name_pat + "*"):
        if file.suffix in (".png", ".svg"):
            continue
        out_name = file.name
        if out_name != name_pat:
            # remove vestigial name from front
            out_name = out_name[len(name_pat) + 1 :]
        cmd = ["dot", f"-T{fmt}", "-o", f"{out_name}.{fmt}", str(file.name)]
        print(" ".join(cmd))
        subprocess.run(cmd, cwd=str(path))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Download, checksum and install CMake
for Linux, Mac and Windows

Automatically determines URL of latest CMake via Git >= 2.18, or manual choice.
"""

from pathlib import Path
import argparse
import subprocess
import re
import sys
import typing as T
import shutil
import urllib.request
import hashlib
import platform
import tarfile
import zipfile
import tempfile

HEAD = "https://github.com/Kitware/CMake/releases/download/"
PLATFORMS = ("amd64", "x86_64", "x64", "i86pc")


def get_latest_version(repo: str, *, tail: str = "", request: str = None) -> str:
    """
    get latest version using Git
    """

    git = shutil.which("git")
    if not git:
        raise FileNotFoundError("Git required to download CMake")

    cmd = [git, "ls-remote", "--tags", "--sort=v:refname", repo]
    revs = subprocess.check_output(cmd, universal_newlines=True).strip().split("\n")
    pat = r".*refs/tags/v(\w+\.\w+\.\w+.*)" + tail

    versions = []
    for v in revs:
        mat = re.match(pat, v)
        if mat:
            versions.append(mat.group(1))

    if request:
        if request in versions:
            return request
        raise ValueError(f"version {request} is not available. Available versions: {versions}")

    return versions[-1]


def url_retrieve(url: str, outfile: Path):
    outfile = Path(outfile).expanduser().resolve()
    if outfile.is_dir():
        raise ValueError("Please specify full filepath, including filename")
    outfile.parent.mkdir(parents=True, exist_ok=True)

    urllib.request.urlretrieve(url, str(outfile))


def file_checksum(fn: Path, hashfn: Path, mode: str) -> bool:
    h = hashlib.new(mode)
    h.update(fn.read_bytes())
    digest = h.hexdigest()

    with hashfn.open("r") as f:
        for line in f:
            if line.startswith(digest):
                if line.split()[-1] == fn.name:
                    return True

    return False


def install_cmake(
    cmake_version: str,
    outfile: Path,
    prefix: Path = None,
    quiet: bool = False,
):

    if sys.platform == "darwin":
        brew = shutil.which("brew")
        if brew:
            subprocess.check_call(["brew", "install", "cmake"])
        else:
            subprocess.check_call(["pip", "install", "cmake"])
        return
    elif sys.platform == "linux":
        if platform.machine().lower() not in PLATFORMS:
            raise ValueError("This method is for Linux 64-bit x86_64 systems")

    prefix = Path(prefix).expanduser().resolve()
    prefix.mkdir(parents=True, exist_ok=True)
    print("Installing CMake to", prefix)

    if outfile.suffix == ".gz":
        with tarfile.open(outfile) as t:
            t.extractall(str(prefix))
        stem = outfile.name.split(".tar.gz")[0]
        # .stem doesn't work as intended for multiple suffixes:  Path("foo.tar.gz").stem == "foo.tar"
    elif outfile.suffix == ".zip":
        with zipfile.ZipFile(outfile) as z:
            z.extractall(str(prefix))
        stem = outfile.stem
    else:
        raise ValueError(f"Unsure how to extract {outfile}")

    if sys.platform == "linux":
        stanza = f"export PATH={prefix / stem}/bin:$PATH"
        for c in ("~/.bashrc", "~/.profile"):
            cfn = Path(c).expanduser()
            if cfn.is_file():
                print("\n\n add to", cfn, stanza)
                break
    elif sys.platform == "win32":
        print(f"add to PATH: {prefix / stem}/bin")
    else:
        raise ValueError(f"Unsure how to install CMake for {sys.platform}")


def cmake_files(cmake_version: str, odir: Path) -> T.Tuple[Path, str]:
    """
    this relies on the per-OS naming scheme used by Kitware in their GitHub Releases
    """

    if sys.platform == "cygwin":
        raise ValueError("use Cygwin setup.exe to install CMake, or manual compile")
    elif sys.platform == "darwin":
        raise ValueError("use brew install cmake, or manual compile")
    elif sys.platform == "linux":
        if platform.machine().lower() not in PLATFORMS:
            raise ValueError("This method is for Linux 64-bit x86_64 systems")
        tail = "linux-x86_64.tar.gz"
    elif sys.platform == "win32":
        tail = "windows-x86_64.zip"
    else:
        raise ValueError(f"unknown platform {sys.platform}")

    ofn = f"cmake-{cmake_version}-{tail}"

    return odir / ofn, HEAD + f"v{cmake_version}/{ofn}"


def download_cmake(outdir: Path, get_version: str) -> Path:

    outdir = Path(outdir).expanduser()
    outdir.mkdir(parents=True, exist_ok=True)
    outfile, url = cmake_files(get_version, outdir)
    # %% checksum
    hashstem = f"cmake-{get_version}-SHA-256.txt"
    hashurl = HEAD + f"v{get_version}/{hashstem}"
    hashfile = outdir / hashstem

    if not hashfile.is_file() or hashfile.stat().st_size == 0:
        url_retrieve(hashurl, hashfile)

    if not outfile.is_file() or outfile.stat().st_size < 1e6:
        url_retrieve(url, outfile)

    if not file_checksum(outfile, hashfile, "sha256"):
        raise ValueError(f"{outfile} SHA256 checksum did not match {hashfile}")

    return outfile


def check_cmake_version(min_version: str) -> bool:
    cmake = shutil.which("cmake")
    if not cmake:
        return False

    cmake_version = subprocess.check_output([cmake, "--version"], universal_newlines=True).split()[2]

    try:
        import pkg_resources

        return pkg_resources.parse_version(cmake_version) >= pkg_resources.parse_version(min_version)
    except ImportError:
        print(f"CMake {cmake_version} already installed.")

    return None


def main():
    p = argparse.ArgumentParser()
    p.add_argument("version", help="request version (default latest)", nargs="?")
    p.add_argument("-o", "--outdir", help="download archive directory", default=tempfile.gettempdir())
    p.add_argument("--prefix", help="Path prefix to install CMake under", default="~/.local")
    p.add_argument("-q", "--quiet", help="non-interactive install", action="store_true")
    p.add_argument("-n", "--dryrun", help="just check version", action="store_true")
    P = p.parse_args()

    get_version = get_latest_version("git://github.com/kitware/cmake.git", tail=r"\^\{\}$", request=P.version)

    if not P.version and check_cmake_version(get_version):
        print(f"You already have the latest CMake {get_version}")
        return

    if P.dryrun:
        print(f"CMake {get_version} is available")
        return

    if sys.platform != "darwin":
        outfile = download_cmake(P.outdir, get_version)
    else:
        outfile = None

    install_cmake(get_version, outfile, P.prefix, P.quiet)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Download and install Ninja
for Linux, Mac and Windows

Automatically determines URL of latest version or manual choice.
"""

import tempfile
from pathlib import Path
from argparse import ArgumentParser, Namespace
import zipfile
import sys
import os
import shutil
import stat
import urllib.request
import subprocess
import pkg_resources
import platform

from .cmake_setup import get_latest_version

HEAD = "https://github.com/ninja-build/ninja/releases/download"
ninja_files = {"win32": "ninja-win.zip", "darwin": "ninja-mac.zip", "linux": "ninja-linux.zip"}
PLATFORMS = ("amd64", "x86_64", "x64", "i86pc")


def main():
    p = ArgumentParser()
    p.add_argument("version", help="request version (default latest)", nargs="?")
    p.add_argument("--prefix", help="Path prefix to install under", default="~/.local/bin")
    p.add_argument("-n", "--dryrun", help="just check version", action="store_true")
    P = p.parse_args()

    cli(P)


def cli(P: Namespace):
    if sys.platform in ("darwin", "linux"):
        if platform.machine().lower() not in PLATFORMS:
            raise ValueError("This method is for Linux 64-bit x86_64 systems")

    outfile = Path(tempfile.gettempdir()) / ninja_files[sys.platform]

    version = get_latest_version("git://github.com/ninja-build/ninja.git", request=P.version)

    if not P.version and check_ninja_version(version):
        print(f"You already have latest Ninja {version}")
        return

    if P.dryrun:
        print(f"Ninja {version} is available")
        return

    url = f"{HEAD}/v{version}/{outfile.name}"

    url_retrieve(url, outfile)

    install_ninja(outfile, P.prefix)


def check_ninja_version(min_version: str) -> bool:
    ninja = shutil.which("ninja")
    if not ninja:
        return False

    ret = subprocess.check_output([ninja, "--version"], universal_newlines=True)

    return pkg_resources.parse_version(ret) >= pkg_resources.parse_version(min_version)


def url_retrieve(url: str, outfile: Path):
    print("downloading", url)
    outfile = Path(outfile).expanduser().resolve()
    if outfile.is_dir():
        raise ValueError("Please specify full filepath, including filename")
    outfile.parent.mkdir(parents=True, exist_ok=True)

    urllib.request.urlretrieve(url, str(outfile))


def install_ninja(outfile: Path, prefix: Path = None):

    prefix = Path(prefix).expanduser().resolve()

    if sys.platform in ("darwin", "linux"):
        if platform.machine().lower() not in PLATFORMS:
            raise ValueError("This method is for Linux 64-bit x86_64 systems")
        stanza = f"export PATH={prefix}:$PATH"
        for c in ("~/.bashrc", "~/.zshrc", "~/.profile"):
            cfn = Path(c).expanduser()
            if cfn.is_file():
                print("\n add to", cfn, "\n\n", stanza)
                break
    else:
        print("add to PATH environment variable:")
        print(prefix)

    # %% extract ninja exe
    prefix.mkdir(parents=True, exist_ok=True)
    print("Installing to", prefix)

    member = "ninja"
    if os.name == "nt":
        member += ".exe"

    with zipfile.ZipFile(outfile) as z:
        z.extract(member, str(prefix))

    os.chmod(prefix / member, stat.S_IRWXU)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
NOTE: most Linux users can simply download and install almost instantly
  instead of this lengthly compilation with cmake_setup.py

---

Does NOT use sudo

Compiles and installs CMake.

Alternative: use binary downloads from:

* pip install cmake
* https://cmake.org/download/

prereqs
CentOS:    yum install gcc-c++ make ncurses-devel openssl-devel unzip
Debian / Ubuntu: apt install g++ make libncurses-dev libssl-dev unzip
Cygwin: setup-x86_64.exe -P gcc-g++ make libncurses-devel libssl-devel

Git > 2.18 required, or specify CMake version at command line e.g.

python cmake_compile.py v3.18.4
"""

import tempfile
import argparse
import os
import subprocess
import tarfile
import shutil
import urllib.request
from pathlib import Path


from .cmake_setup import get_latest_version, file_checksum

url_stem = "https://github.com/Kitware/CMake/releases/download"


def main():
    p = argparse.ArgumentParser()
    p.add_argument("version", nargs="?")
    p.add_argument("-prefix", help="where to install CMake")
    p.add_argument("-workdir", help="use existing source code path")
    p = p.parse_args()

    # 1. download
    if p.workdir:
        src_root = Path(p.workdir).expanduser()
    else:
        # get latest CMake version if not specified
        version = get_latest_version("git://github.com/kitware/cmake.git", tail=r"\^\{\}$", request=p.version)

        src_root = download_and_extract(tempfile.gettempdir(), version)

    prefix = None
    if p.prefix:
        prefix = Path(p.prefix).expanduser()
        print("installing cmake:", prefix)
        prefix.mkdir(parents=True, exist_ok=True)

    if shutil.which("cmake"):
        cmake_build(src_root, prefix)
    else:
        bootstrap(src_root, prefix)

    print("\n----------------------------------------------------")
    if prefix:
        print(f"please add to your system PATH:  {prefix}/bin/")
        print("\nreopen a new terminal to use CMake", version)
    else:
        print(f"CMake built under {src_root}")
        print("To install CMake, rerun using cmake_compile -prefix option set to desired install path.")


def job_count() -> int:
    """
    < 2GB RAM is a problem for parallel builds
    Make is more problematic than Ninja in this regard
    if ram or cpu count cannot be determined, return 1
    """

    Njobs = 1
    maxcpu = None
    try:
        import psutil

        Ncpu = psutil.cpu_count(logical=False)
        avail = psutil.virtual_memory().available
        if avail < 1e9:
            maxcpu = 2
        elif avail < 1.5e9:
            maxcpu = 4
        else:
            maxcpu = Ncpu
    except ImportError:
        pass

    if maxcpu is not None and Ncpu is not None:
        # we know free ram and CPU count
        Njobs = min(maxcpu, Ncpu)

    return Njobs


def cmake_build(src_root: Path, prefix: Path):

    build_root = src_root / "build"
    build_root.mkdir(parents=True, exist_ok=True)

    opts = [
        "-DCMAKE_BUILD_TYPE=Release",
        "-DCMAKE_USE_OPENSSL:BOOL=ON",
        "-DBUILD_TESTING:BOOL=OFF",
    ]
    if prefix:
        opts.append(f"-DCMAKE_INSTALL_PREFIX:PATH={prefix}")
    if os.environ.get("CMAKE_GENERATOR") and os.environ["CMAKE_GENERATOR"] == "Ninja":
        opts.append("-GNinja")

    cmd = ["cmake", str(src_root)] + opts
    print(" ".join(cmd))
    subprocess.check_call(cmd, cwd=build_root)

    popts = ["--parallel"]
    Njobs = job_count()
    if Njobs:
        popts.append(str(Njobs))
    cmd = ["cmake", "--build", str(build_root)] + popts
    print(" ".join(cmd))
    subprocess.check_call(cmd)

    if prefix:
        print("installing cmake:", prefix)
        cmd = ["cmake", "--build", str(build_root), "--target", "install"]
        print(" ".join(cmd))
        subprocess.check_call(cmd)


def bootstrap(src_root: Path, prefix: Path):
    """ cmake Unix bootstrap """

    if os.name == "nt":
        raise RuntimeError("CMake bootstrap is for Unix-like systems only")

    cmake_bootstrap = src_root / "bootstrap"
    if not cmake_bootstrap.is_file():
        raise FileNotFoundError(cmake_bootstrap)

    opts = [
        "--",
        "-DCMAKE_BUILD_TYPE=Release",
        "-DCMAKE_USE_OPENSSL:BOOL=ON",
    ]

    if prefix:
        opts.append(f"--prefix={prefix}")

    Njobs = job_count()
    if Njobs:
        opts.append(f"--parallel={Njobs}")

    cmd = [str(cmake_bootstrap)] + opts
    print(" ".join(cmd))
    subprocess.check_call(cmd, cwd=src_root)

    popts = ["-j"]
    if Njobs:
        popts.append(str(Njobs))

    cmd = ["make"] + popts
    print(" ".join(cmd))
    subprocess.check_call(cmd, cwd=src_root)

    if prefix:
        print("installing cmake:", prefix)
        subprocess.check_call(["make", "install"], cwd=src_root)


def download_and_extract(workdir: Path, version: str) -> Path:

    workdir = Path(workdir)
    workdir.mkdir(exist_ok=True)
    stem = f"cmake-{version}"
    src_root = (workdir / stem).resolve(strict=False)
    cfn = f"{stem}-SHA-256.txt"
    fn = f"{stem}.tar.gz"

    cmake_archive = workdir / fn
    url = f"{url_stem}/v{version}/{fn}"
    if not cmake_archive.is_file():
        print("downloading", url)
        urllib.request.urlretrieve(url, cmake_archive)

    # 2. build
    cmake_sig = workdir / cfn
    url_sig = f"{url_stem}/v{version}/{cfn}"
    if not cmake_sig.is_file():
        print("downloading", url_sig)
        urllib.request.urlretrieve(url_sig, cmake_sig)

    print("checking SHA256 signature")
    if not file_checksum(cmake_archive, cmake_sig, "sha256"):
        raise ValueError(f"{cmake_archive} SHA256 checksum did not match {cmake_sig}")

    if not src_root.is_dir():
        print("extracting CMake source")
        with tarfile.open(str(cmake_archive)) as tf:
            tf.extractall(str(workdir))

    return src_root


if __name__ == "__main__":
    main()
